/* ssh-waker, by mulix for NiGhTHiNG*/
/* cobbled together from various sources, only the glue is original */
/* the wake on lan is copied verbatim from ether-wake.c */

#include <libipq/libipq.h>
#include <linux/netfilter.h>
#include <sys/ioctl.h>


#include "util.h"

#define IPQ_BUF_SIZE 10240

static char* buf;
struct ipq_handle* qh;
static char mac_address[] = "00:11:22:33:44:55";
static char ifname[] = "eth0";
unsigned char outpack[1000];

static int 
get_fill(unsigned char *pkt, char *arg)
{
	int sa[6];
	unsigned char station_addr[6];
	int byte_cnt;
	int offset, i;
	char *cp;

	for (cp = arg; *cp; cp++)
		if (*cp != ':' && !isxdigit(*cp)) {
			warn("ping: patterns must be specified as hex digits.\n");
			return -1;
		}

	byte_cnt = sscanf(arg, "%2x:%2x:%2x:%2x:%2x:%2x",
					  &sa[0], &sa[1], &sa[2], &sa[3], &sa[4], &sa[5]);
	for (i = 0; i < 6; i++)
		station_addr[i] = sa[i];

	log("Command line stations address is "
		"%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x.\n",
		sa[0], sa[1], sa[2], sa[3], sa[4], sa[5]);

	if (byte_cnt != 6) {
		warn("ping: The Magic Packet address must be specified as "
			 "00:11:22:33:44:55.\n");
		return -1;
	}
	
	memcpy(pkt, station_addr, 6);
	memcpy(pkt+6, station_addr, 6);
	pkt[12] = 0x08;				/* Or 0x0806 for ARP, 0x8035 for RARP */
	pkt[13] = 0x42;
	offset = 14;
	
	memset(pkt+offset, 0xff, 6);
	offset += 6;

	for (i = 0; i < 16; i++) {
		memcpy(pkt+offset, station_addr, 6);
		offset += 6;
	}

	return offset;
}




static void
send_wake_on_lan_packet()
{
	int pktsize;
	int s, i;
	struct ifreq if_hwaddr;
	unsigned char *hwaddr = if_hwaddr.ifr_hwaddr.sa_data;
	struct sockaddr whereto;

	/* Note: PF_INET, SOCK_DGRAM, IPPROTO_UDP would allow SIOCGIFHWADDR to
	   work as non-root, but we need SOCK_PACKET to specify the Ethernet
	   destination address. */
	if ((s = socket(AF_INET, SOCK_PACKET, SOCK_PACKET)) < 0) {
		warn("failed, error is '%s'", strerror(errno));
	}
	
	/* Don't revert if debugging allows a normal user to get the raw socket. */
	setuid(getuid());

	pktsize = get_fill(outpack, mac_address);
	if (pktsize == -1){
		warn("failed to create packet, aborting WakeOnLan");
		return;
	}

	/* Fill in the source address, if possible.
	   The code to retrieve the local station address is Linux specific. */
	strcpy(if_hwaddr.ifr_name, ifname);
	if (ioctl(s, SIOCGIFHWADDR, &if_hwaddr) < 0) {
		warn("SIOCGIFHWADDR on %s failed: %s\n", ifname,
			 strerror(errno));
		return;
	}

	memcpy(outpack+6, if_hwaddr.ifr_hwaddr.sa_data, 6);

	log("The hardware address (SIOCGIFHWADDR) of %s is type %d  "
		"%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x.\n", ifname,
		if_hwaddr.ifr_hwaddr.sa_family, hwaddr[0], hwaddr[1],
		hwaddr[2], hwaddr[3], hwaddr[4], hwaddr[5]);
	
	
	whereto.sa_family = 0;
	strcpy(whereto.sa_data, ifname);
	
	if ((i = sendto(s, outpack, pktsize, 0, &whereto, sizeof(whereto))) < 0)
		warn("sendto, error is '%s'", strerror(errno));
	else 
		log("Sendto worked ! returned %d.\n", i);
}


static void
run_main_netfilter_loop()
{
    ssize_t len;

    while (1) {
		int ptype,error;
		ipq_packet_msg_t* packet;
		
		len = ipq_read(qh, buf, IPQ_BUF_SIZE, 0);
		if (len < 0) {
			log("returned '%d', error is '%s'",
				len, ipq_errstr());
			continue;
		} else if (len == 0) {
			log("timeout exceeded");
			continue;
		} else {
			ptype = ipq_message_type(buf);
			packet = ipq_get_packet(buf);
			log ("received packet, length=%d, type=%d\n",
				 len, ptype);
			switch (ptype) {
			case NLMSG_ERROR:
				error = -ipq_get_msgerr(buf);
				log ("NLMSG ERROR: %d", error);
				break;
			case IPQM_PACKET:
				log("got a packet, sending WakeOnLan packet");
				send_wake_on_lan_packet();
				log("sleeping");
				sleep(1); 
				log("calling ipq_set_verdict");
				error = ipq_set_verdict(qh, packet->packet_id, NF_ACCEPT, 0, NULL);
				if (error == -1){
				     log("ipq_set_verdict failed, error '%s'", ipq_errstr());
				}else{
				     log("ipq_set_verdict succesfull!");
				}
				break;
			default:
				log("got unknown packet type '%d'", ptype);
				break;
			}
		}
		continue;
	}
}




static int
hook_into_netfilter()
{
	qh = ipq_create_handle(0);
	if (!qh){
		warn("failed to create ipq handle, error is '%s'",
			 ipq_errstr());
		return 0;
	}
	
	if (ipq_set_mode(qh, IPQ_COPY_PACKET, 0xFFFF) < 0) {
		warn("can't set netlink mode for packet coping, error is '%s'",
			 ipq_errstr());
		return 0;
	}
	
	buf = (char *)malloc(IPQ_BUF_SIZE);
	if (!buf) {
		warn("could not allocate buffer");
		return 0;
	}
	
	return 1;
}


int 
main(int argc, char** argv)
{
	/* hook into the netfilter QUEUE target */
	if (!hook_into_netfilter()){
		fatal("could not hook into netfilter");
	}
	
	log("running main netfilter loop");
	run_main_netfilter_loop();
}


